from dataclasses import dataclass
from typing import Optional, Sequence

from . import html, outcome
from .core import Link, Url


@dataclass(frozen=True)
class LinkExtractor(outcome.Converter[Optional[Sequence[Link]]]):
    url: Url

    def convert_redirect(self, redirect: outcome.Redirect) -> None:
        return None

    def convert_page(self, page: outcome.Page) -> Sequence[Link]:
        return html.get_links(body=page.body, url=self.url)

    def convert_request_error(self, error: outcome.RequestError) -> None:
        return None

    def convert_excluded(self, excluded: outcome.Excluded) -> None:
        return None

    def convert_unknown(self, unknown: outcome.Unknown) -> None:
        return None


def get_links(url: Url, result: outcome.Result) -> Optional[Sequence[Link]]:
    converter = LinkExtractor(url=url)
    return result.convert_with(converter)



import asyncio
from typing import AbstractSet

from .core import Url
from .link_extractor import get_links
from .monitor import Monitor
from .requester import Requester
from .url_store import UrlInfo, UrlStore


async def investigate_url(
    requester: Requester,
    url_store: UrlStore,
    start_url: Url,
    url: Url,
) -> AbstractSet[Url]:
    """
    Follow HTTP link and return new links if any are found.

    For external websites this only does a `HEAD` to know if the link is broken or not, so
    no new links are returned.
    """

    if url.netloc != start_url.netloc:
        result = await requester.get(url=url, use_head=True)

        if result.status_code() == 405:  # method not allowed
            result = await requester.get(url=url)

        return url_store.add_page(
            url=url,
            info=UrlInfo(result=result, links=None),
        )

    result = await requester.get(url=url)

    return url_store.add_page(
        url=url,
        info=UrlInfo(
            result=result,
            links=(get_links(url=url, result=result) if result.ok() else None),
        ),
    )


async def work(
    queue: asyncio.Queue[Url],
    requester: Requester,
    url_store: UrlStore,
    monitor: Monitor,
    start_url: Url,
):
    while True:
        task_url = await queue.get()
        monitor.on_task_start(queued=queue.qsize())
        try:
            new_urls = await investigate_url(
                requester=requester,
                url_store=url_store,
                start_url=start_url,
                url=task_url,
            )
            for url in new_urls:
                queue.put_nowait(url)
        finally:
            queue.task_done()

        monitor.on_task_done(
            queued=queue.qsize(),
            result=url_store.get_url_infos()[task_url].result,
        )



import json
import subprocess

from flask import Blueprint, redirect

from . import util


def make_blueprint() -> Blueprint:
    blueprint = Blueprint("main", __name__)

    @blueprint.route("/")
    def root():
        return """
            <a href="/foo">
            <a href="/bar">
        """

    @blueprint.route("/foo")
    def foo():
        return redirect("/baz")

    @blueprint.route("/bar")
    def bar():
        return redirect("/nx")

    @blueprint.route("/baz")
    def baz():
        return """<a href="/">\n"""

    return blueprint


def test_text(http_server) -> None:
    http_server(blueprint=make_blueprint(), port=5000)

    result = subprocess.run(
        util.command(url="http://localhost:5000"),
        stdout=subprocess.PIPE,
    )

    assert result.returncode == 1
    assert result.stdout.decode() == util.output_str(
        """
        📂 Results: 5 links (3 ok, 2 failed)
        ├── 📄 http://localhost:5000
        │   └── 🔗 /bar: 302 → 404
        └── 📄 http://localhost:5000/
            └── 🔗 /bar: 302 → 404
        """
    )


def test_json(http_server) -> None:
    http_server(blueprint=make_blueprint(), port=5000)

    result = subprocess.run(
        util.command(url="http://localhost:5000", json=True),
        stdout=subprocess.PIPE,
    )

    assert result.returncode == 1
    assert json.loads(result.stdout.decode()) == {
        "http://localhost:5000": {
            "links": [
                {
                    "href": "/foo",
                    "url": "http://localhost:5000/foo",
                    "results": [
                        {
                            "type": "redirect",
                            "status_code": 302,
                            "value": "/baz",
                            "url": "http://localhost:5000/baz",
                        },
                        {
                            "type": "response",
                            "status_code": 200,
                        },
                    ],
                },
                {
                    "href": "/bar",
                    "url": "http://localhost:5000/bar",
                    "results": [
                        {
                            "type": "redirect",
                            "status_code": 302,
                            "value": "/nx",
                            "url": "http://localhost:5000/nx",
                        },
                        {
                            "type": "response",
                            "status_code": 404,
                        },
                    ],
                },
            ],
        },
        "http://localhost:5000/": {
            "links": [
                {
                    "href": "/foo",
                    "url": "http://localhost:5000/foo",
                    "results": [
                        {
                            "type": "redirect",
                            "status_code": 302,
                            "value": "/baz",
                            "url": "http://localhost:5000/baz",
                        },
                        {
                            "type": "response",
                            "status_code": 200,
                        },
                    ],
                },
                {
                    "href": "/bar",
                    "url": "http://localhost:5000/bar",
                    "results": [
                        {
                            "type": "redirect",
                            "status_code": 302,
                            "value": "/nx",
                            "url": "http://localhost:5000/nx",
                        },
                        {
                            "type": "response",
                            "status_code": 404,
                        },
                    ],
                },
            ],
        },
        "http://localhost:5000/baz": {
            "links": [
                {
                    "href": "/",
                    "url": "http://localhost:5000/",
                    "results": [
                        {
                            "type": "response",
                            "status_code": 200,
                        },
                    ],
                },
            ],
        },
    }



from dataclasses import dataclass
from urllib.parse import urldefrag, urlparse


@dataclass(frozen=True)
class Url:
    """
    Wrapper around URL strings.

    Use `Url.from_str` to build an instance and `url.full` to get the underlying string
    (e.g. for communicating with HTTP libraries).
    """

    full: str
    scheme: str
    netloc: str

    @classmethod
    def from_str(cls, url: str) -> "Url":
        parsed = urlparse(url)
        assert parsed.scheme, f"Invalid URL: {url}"
        assert parsed.netloc, f"Invalid URL: {url}"
        return cls(
            full=urldefrag(url).url,
            scheme=parsed.scheme,
            netloc=parsed.netloc,
        )

    def __str__(self) -> str:
        return self.full


@dataclass(frozen=True)
class Link:
    href: str
    url: Url



import json
from dataclasses import dataclass
from typing import Any, Mapping, Sequence

from . import outcome
from .analyzer import Analysis, LinkResult, Page
from .core import Url


@dataclass(frozen=True)
class Converter(outcome.Converter[Any]):
    def convert_page(self, page: outcome.Page) -> Any:
        return {
            "type": "response",
            "status_code": page.code,
        }

    def convert_redirect(self, redirect: outcome.Redirect) -> Any:
        return {
            "type": "redirect",
            "status_code": redirect.code,
            "value": redirect.ref,
            "url": str(redirect.url),
        }

    def convert_request_error(self, error: outcome.RequestError) -> Any:
        return {
            "type": "request_error",
            "message": str(error.msg),
        }

    def convert_excluded(self, excluded: outcome.Excluded) -> Any:
        return {
            "type": "excluded",
        }

    def convert_unknown(self, unknown: outcome.Unknown) -> Any:
        return {
            "type": "unknown",
        }


def results_to_json(results: outcome.Results) -> Sequence[Any]:
    return results.convert_with(Converter())


def link_to_json(link: LinkResult) -> Any:
    return {
        "href": link.href,
        "url": str(link.url),
        "results": results_to_json(link.results),
    }


def page_to_json(page: Page) -> Any:
    return {
        "links": [link_to_json(link) for link in page.links],
    }


def pages_to_json(pages: Mapping[Url, Page]) -> Any:
    return {url.full: page_to_json(page) for (url, page) in pages.items()}


def dump_json(analysis: Analysis) -> str:
    obj = pages_to_json(pages=analysis.pages)
    return json.dumps(obj)



import pytest

from discolinks.core import Url
from discolinks.excluder import Excluder, ExcluderRegexError


def test_excluder_from_regexes_error():
    with pytest.raises(ExcluderRegexError) as error:
        Excluder.from_regexes("(")

    assert error.value.regex == "("
    assert error.value.msg == "missing ), unterminated subpattern at position 0"


@pytest.mark.parametrize(
    "regexes,url,expected",
    [
        ([], Url.from_str("https://foo"), False),
        (["a"], Url.from_str("https://foo"), False),
        (["f"], Url.from_str("https://foo"), True),
        (["^f"], Url.from_str("https://foo"), False),
        (["^h"], Url.from_str("https://foo"), True),
        (["a", "b"], Url.from_str("https://foo"), False),
        (["a", "f"], Url.from_str("https://foo"), True),
    ],
)
def test_excluder_combined(regexes: str, url: Url, expected: bool):
    result = Excluder.from_regexes(regexes).is_excluded(url)

    assert result is expected



import subprocess

import pytest


@pytest.mark.parametrize(
    "invocation",
    [
        ["discolinks"],
        ["python", "-m", "discolinks"],
    ],
)
def test_version_module(invocation):
    result = subprocess.run(
        [*invocation, "--version"],
        stdout=subprocess.PIPE,
    )

    assert result.returncode == 0
    assert result.stdout.decode().startswith("discolinks version ")



import json
import subprocess

from flask import Blueprint, abort

from . import util


def make_blueprint_with_link(link: str) -> Blueprint:
    blueprint = Blueprint("main", __name__)

    @blueprint.route("/")
    def root():
        return f"""<a href="{link}">\n"""

    return blueprint


def make_blueprint_no_head() -> Blueprint:
    blueprint = Blueprint("main", __name__)

    @blueprint.route("/", methods=["HEAD"])
    def root_head():
        abort(405)

    @blueprint.route("/", methods=["GET"])
    def root_get():
        return ""

    return blueprint


def test_json(http_server) -> None:
    http_server(
        blueprint=make_blueprint_with_link(link="http://localhost:5001"),
        port=5000,
    )
    http_server(blueprint=make_blueprint_no_head(), port=5001)

    result = subprocess.run(
        util.command(url="http://localhost:5000", json=True),
        stdout=subprocess.PIPE,
    )

    assert result.returncode == 0
    assert json.loads(result.stdout.decode()) == {
        "http://localhost:5000": {
            "links": [
                {
                    "href": "http://localhost:5001",
                    "url": "http://localhost:5001",
                    "results": [
                        {
                            "type": "response",
                            "status_code": 200,
                        }
                    ],
                },
            ],
        },
    }



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, Optional, Sequence, TypeVar

from .core import Url

Out = TypeVar("Out")


@dataclass(frozen=True)
class Result(ABC):
    @abstractmethod
    def ok(self) -> bool:
        pass

    @abstractmethod
    def status_code(self) -> Optional[int]:
        pass

    @abstractmethod
    def redirect_url(self) -> Optional[Url]:
        pass

    @abstractmethod
    def error_msg(self) -> Optional[str]:
        pass

    @abstractmethod
    def convert_with(self, converter: "Converter[Out]") -> Out:
        pass


@dataclass(frozen=True)
class Redirect(Result):
    code: int
    ref: str
    url: Url

    def ok(self) -> bool:
        return True

    def status_code(self) -> Optional[int]:
        return self.code

    def redirect_url(self) -> Optional[Url]:
        return self.url

    def error_msg(self) -> Optional[str]:
        return None

    def convert_with(self, converter: "Converter[Out]") -> Out:
        return converter.convert_redirect(self)


@dataclass(frozen=True)
class Page(Result):
    code: int
    body: str

    def ok(self) -> bool:
        return not (400 <= self.code < 600)

    def status_code(self) -> Optional[int]:
        return self.code

    def redirect_url(self) -> Optional[Url]:
        return None

    def error_msg(self) -> Optional[str]:
        return None

    def convert_with(self, converter: "Converter[Out]") -> Out:
        return converter.convert_page(self)


@dataclass(frozen=True)
class RequestError(Result):
    msg: str

    def ok(self) -> bool:
        return False

    def status_code(self) -> Optional[int]:
        return None

    def redirect_url(self) -> Optional[Url]:
        return None

    def error_msg(self) -> Optional[str]:
        return self.msg

    def convert_with(self, converter: "Converter[Out]") -> Out:
        return converter.convert_request_error(self)


@dataclass(frozen=True)
class Excluded(Result):
    def ok(self) -> bool:
        return True

    def status_code(self) -> Optional[int]:
        return None

    def redirect_url(self) -> Optional[Url]:
        return None

    def error_msg(self) -> Optional[str]:
        return None

    def convert_with(self, converter: "Converter[Out]") -> Out:
        return converter.convert_excluded(self)


@dataclass(frozen=True)
class Unknown(Result):
    def ok(self) -> bool:
        # It's considered OK because the underlying error will be reported by another
        # mechanism.
        return True

    def status_code(self) -> Optional[int]:
        return None

    def redirect_url(self) -> Optional[Url]:
        return None

    def error_msg(self) -> Optional[str]:
        return None

    def convert_with(self, converter: "Converter[Out]") -> Out:
        return converter.convert_unknown(self)


@dataclass(frozen=True)
class Results:
    chain: Sequence[Result]

    def ok(self) -> bool:
        final = self.chain[-1]
        return final.ok()

    def convert_with(self, converter: "Converter[Out]") -> Sequence[Out]:
        return tuple(item.convert_with(converter) for item in self.chain)


class Converter(ABC, Generic[Out]):
    @abstractmethod
    def convert_redirect(self, redirect: Redirect) -> Out:
        pass

    @abstractmethod
    def convert_page(self, page: Page) -> Out:
        pass

    @abstractmethod
    def convert_request_error(self, error: RequestError) -> Out:
        pass

    @abstractmethod
    def convert_excluded(self, excluded: Excluded) -> Out:
        pass

    @abstractmethod
    def convert_unknown(self, unknown: Unknown) -> Out:
        pass



from dataclasses import dataclass, field
from typing import Mapping, Optional, Sequence

from . import outcome
from .core import Link, Url


@dataclass(frozen=True)
class UrlInfo:
    result: outcome.Result
    links: Optional[Sequence[Link]]

    def link_urls(self) -> frozenset[Url]:
        redirect_url = self.result.redirect_url()

        if redirect_url is not None:
            return frozenset([redirect_url])
        elif self.links is None:
            return frozenset()
        else:
            return frozenset(link.url for link in self.links)


@dataclass(frozen=True)
class UrlStore:
    url_infos: dict[Url, UrlInfo] = field(init=False, default_factory=dict)
    seen_urls: set[Url] = field(init=False, default_factory=set)

    def add_page(self, url: Url, info: UrlInfo) -> frozenset[Url]:
        """
        Store page information for a given URL and return new URLs.

        This can only be called once for each URL and each discovered URL is only returned
        once.
        """

        assert url not in self.url_infos, f"URL already stored: {url}"
        self.url_infos[url] = info
        self.seen_urls.add(url)
        new_urls = info.link_urls() - self.seen_urls
        self.seen_urls.update(new_urls)
        return new_urls

    def count(self) -> int:
        return len(self.seen_urls)

    def get_url_infos(self) -> Mapping[Url, UrlInfo]:
        return self.url_infos



import logging
import ssl
from dataclasses import dataclass, field
from typing import Union

import httpx

from . import outcome
from .core import Url
from .excluder import Excluder

logger = logging.getLogger(__name__)


def httpx_to_result(response: httpx.Response) -> outcome.Result:
    if response.next_request is not None:
        return outcome.Redirect(
            code=response.status_code,
            ref=response.headers["location"],
            url=Url.from_str(str(response.next_request.url)),
        )
    else:
        return outcome.Page(
            code=response.status_code,
            body=response.text,
        )


def httpx_to_error(error: Union[httpx.RequestError, ssl.SSLError]) -> str:
    if isinstance(error, httpx.TimeoutException):
        return "Network timeout"

    return str(error)


@dataclass(frozen=True)
class Requester:
    excluder: Excluder
    client: httpx.AsyncClient = field(init=False, default_factory=httpx.AsyncClient)

    async def get(self, url: Url, use_head: bool = False) -> outcome.Result:
        """
        Fetch a page from the given HTTP URL.
        """
        method = "HEAD" if use_head else "GET"

        if self.excluder.is_excluded(url):
            logger.debug("Excluded: %s", url)
            return outcome.Excluded()
        else:
            logger.debug("%s %s", method, url)

        try:
            response = await self.client.request(method=method, url=url.full)
        except (httpx.RequestError, ssl.SSLError) as error:
            msg = httpx_to_error(error)
            return outcome.RequestError(msg=msg)

        return httpx_to_result(response)



import subprocess

from flask import Blueprint, redirect

from . import util


def make_blueprint() -> Blueprint:
    blueprint = Blueprint("main", __name__)

    @blueprint.route("/")
    def root():
        return redirect("/foo")

    @blueprint.route("/foo")
    def foo():
        return redirect("/")

    return blueprint


def test_json(http_server) -> None:
    http_server(blueprint=make_blueprint(), port=5000)

    result = subprocess.run(
        util.command(
            url="http://localhost:5000",
            json=True,
        ),
        stdout=subprocess.PIPE,
    )

    assert result.returncode == 1
    assert result.stdout.decode() == ""



import json
import subprocess

import pytest
from flask import Blueprint

from . import util


def make_blueprint() -> Blueprint:
    blueprint = Blueprint("main", __name__)

    @blueprint.route("/")
    def root():
        return """<a href="/foo">\n"""

    @blueprint.route("/foo")
    def foo():
        return """<a href="/">\n"""

    return blueprint


def test_text(http_server) -> None:
    http_server(blueprint=make_blueprint(), port=5000)

    result = subprocess.run(
        util.command(url="http://localhost:5000"),
        stdout=subprocess.PIPE,
    )

    assert result.returncode == 0
    assert result.stdout.decode() == util.output_str(
        "📂 Results: 3 links (3 ok, 0 failed)"
    )


@pytest.mark.parametrize(
    "max_parallel_requests",
    [
        None,
        1,
        4,
    ],
)
def test_json(max_parallel_requests: int, http_server) -> None:
    http_server(blueprint=make_blueprint(), port=5000)

    result = subprocess.run(
        util.command(
            url="http://localhost:5000",
            json=True,
            max_parallel_requests=max_parallel_requests,
        ),
        stdout=subprocess.PIPE,
    )

    assert result.returncode == 0
    assert json.loads(result.stdout.decode()) == {
        "http://localhost:5000": {
            "links": [
                {
                    "href": "/foo",
                    "url": "http://localhost:5000/foo",
                    "results": [
                        {
                            "type": "response",
                            "status_code": 200,
                        }
                    ],
                },
            ],
        },
        "http://localhost:5000/": {
            "links": [
                {
                    "href": "/foo",
                    "url": "http://localhost:5000/foo",
                    "results": [
                        {
                            "type": "response",
                            "status_code": 200,
                        }
                    ],
                },
            ],
        },
        "http://localhost:5000/foo": {
            "links": [
                {
                    "href": "/",
                    "url": "http://localhost:5000/",
                    "results": [
                        {
                            "type": "response",
                            "status_code": 200,
                        }
                    ],
                },
            ],
        },
    }



import json
import subprocess

from flask import Blueprint

from . import util


def make_blueprint() -> Blueprint:
    blueprint = Blueprint("main", __name__)

    @blueprint.route("/")
    def root():
        return """
            <a href="/foo"></a>
        """

    @blueprint.route("/foo")
    def foo():
        return """
            <a href="/nonexistent"></a>
        """

    return blueprint


def test_ok(http_server) -> None:
    http_server(make_blueprint(), port=5000)

    result = subprocess.run(
        util.command(url="http://localhost:5000", exclude=["foo"], json=True),
        stdout=subprocess.PIPE,
    )

    assert result.returncode == 0
    assert json.loads(result.stdout.decode()) == {
        "http://localhost:5000": {
            "links": [
                {
                    "href": "/foo",
                    "url": "http://localhost:5000/foo",
                    "results": [
                        {
                            "type": "excluded",
                        },
                    ],
                },
            ],
        },
    }


def test_bad_regex(http_server) -> None:
    http_server(make_blueprint(), port=5000)

    result = subprocess.run(
        util.command(url="http://localhost:5000", exclude=["("], json=True),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    assert result.returncode == 1
    assert result.stdout.decode() == ""



import inspect
from typing import Optional, Sequence


def output_str(s: str) -> str:
    clean_str = inspect.cleandoc(s)
    return f"{clean_str}\n"


def command(
    url: str,
    verbose: Optional[bool] = None,
    json: Optional[bool] = None,
    exclude: Sequence[str] = (),
    max_parallel_requests: Optional[int] = None,
) -> Sequence[str]:
    """
    Generate command-line strings based on function parameters.

    The parameters and their default values are as close to the real CLI as possible, so
    that the function call looks pretty much the same as a CLI invocation.
    """

    cli = ["discolinks"]

    if verbose:
        cli += ["--verbose"]

    if json:
        cli += ["--json"]

    for s in exclude:
        cli += ["--exclude", s]

    if max_parallel_requests is not None:
        cli += ["--max-parallel-requests", str(max_parallel_requests)]

    cli += ["--url", url]

    return cli



import subprocess

from flask import Blueprint

from . import util


def make_blueprint() -> Blueprint:
    return Blueprint("main", __name__)


def test(http_server) -> None:
    http_server(blueprint=make_blueprint(), port=5000)

    result = subprocess.run(
        util.command(url="http://localhost:5000", json=True),
        stdout=subprocess.PIPE,
    )

    assert result.returncode == 1
    assert result.stdout.decode() == ""



from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import Iterator

import rich.console
import rich.status

from . import outcome


@dataclass
class Stats:
    queued: int = 0
    in_progress: int = 0
    finished: int = 0
    ok: int = 0
    failed: int = 0

    def on_task_started(self, queued: int) -> None:
        self.queued = queued
        self.in_progress += 1

    def on_task_done(self, queued: int, result: outcome.Result) -> None:
        self.queued = queued
        self.in_progress -= 1
        self.finished += 1
        if result.ok():
            self.ok += 1
        else:
            self.failed += 1


@dataclass(frozen=True)
class Monitor:
    """
    Handle updates of the status bar during scraping.

    Create an instance with `start` and finish the execution with `stop`.
    """

    console: rich.console.Console
    status: rich.status.Status
    stats: Stats = field(init=False, default_factory=Stats)

    @classmethod
    def start(cls, console: rich.console.Console) -> "Monitor":
        status = rich.status.Status(status="Working", console=console)
        status.start()
        return cls(console=console, status=status)

    def print(self, msg: str) -> None:
        self.console.print(msg, markup=False, emoji=False)

    def _update_status(self) -> None:
        self.status.update(
            f"Working:"
            f" [bold blue]{self.stats.queued} [dim white]queued"
            f" → [not dim][bold yellow]{self.stats.in_progress} [dim white]in progress"
            f" → [not dim][bold white]{self.stats.finished} [dim white]finished"
            f" ([not dim][bold green]{self.stats.ok} [dim white]ok,"
            f" [not dim][bold red]{self.stats.failed} [dim white]failed)"
        )

    def on_task_start(self, queued: int) -> None:
        self.stats.on_task_started(queued=queued)
        self._update_status()

    def on_task_done(self, result: outcome.Result, queued: int) -> None:
        self.stats.on_task_done(queued=queued, result=result)
        self._update_status()

    def stop(self) -> None:
        self.status.stop()


@contextmanager
def new_monitor(console: rich.console.Console) -> Iterator[Monitor]:
    """
    Yield a `Monitor` instance and ensure it is stopped properly.
    """

    monitor = Monitor.start(console=console)
    try:
        yield monitor
    finally:
        monitor.stop()



import json
import subprocess

import pytest
from flask import Blueprint

from . import util


def make_blueprint() -> Blueprint:
    blueprint = Blueprint("main", __name__)

    @blueprint.route("/")
    def root():
        return """
            <a href="/foo#bar">
            <a href="/foo#baz">
        """

    @blueprint.route("/foo")
    def foo():
        return """<h2 id="bar">"""

    return blueprint


@pytest.mark.parametrize(
    "url",
    [
        "http://localhost:5000",
        "http://localhost:5000#nx",
    ],
)
def test_json(http_server, url) -> None:
    http_server(blueprint=make_blueprint(), port=5000)

    result = subprocess.run(
        util.command(
            url=url,
            json=True,
        ),
        stdout=subprocess.PIPE,
    )

    assert result.returncode == 0
    assert json.loads(result.stdout.decode()) == {
        "http://localhost:5000": {
            "links": [
                {
                    "href": "/foo#bar",
                    "url": "http://localhost:5000/foo",
                    "results": [
                        {
                            "type": "response",
                            "status_code": 200,
                        }
                    ],
                },
                {
                    "href": "/foo#baz",
                    "url": "http://localhost:5000/foo",
                    "results": [
                        {
                            "type": "response",
                            "status_code": 200,
                        }
                    ],
                },
            ],
        },
        "http://localhost:5000/foo": {
            "links": [],
        },
    }



import re
from dataclasses import dataclass
from typing import Sequence

from .core import Url


@dataclass(frozen=True)
class ExcluderRegexError(Exception):
    regex: str
    msg: str


def parse_regex(regex: str) -> re.Pattern:
    try:
        return re.compile(regex)
    except re.error as error:
        raise ExcluderRegexError(regex=regex, msg=str(error)) from error


@dataclass(frozen=True)
class Excluder:
    """
    Preprocessed engine excluding certain URLs
    """

    patterns: Sequence[re.Pattern]

    @classmethod
    def from_regexes(cls, regexes: Sequence[str]) -> "Excluder":
        """
        Parse regular expressions to be used as exclusion patterns.

        Raises `ExcluderRegexError` if any of the regexes can't be parsed.
        """
        patterns = [parse_regex(regex) for regex in regexes]
        return cls(patterns=patterns)

    def is_excluded(self, url: Url) -> bool:
        return any(re.search(pattern, str(url)) for pattern in self.patterns)



import json
import subprocess

from flask import Blueprint, abort, redirect

from . import util


def make_blueprint(last_redirect_ok: bool) -> Blueprint:
    blueprint = Blueprint("main", __name__)

    @blueprint.route("/")
    def root():
        return redirect("/foo")

    @blueprint.route("/foo")
    def foo():
        return redirect("/bar")

    @blueprint.route("/bar")
    def bar():
        if last_redirect_ok:
            return """<a href="/">\n"""
        else:
            abort(404)

    return blueprint


def test_json_ok(http_server) -> None:
    http_server(blueprint=make_blueprint(last_redirect_ok=True), port=5000)

    result = subprocess.run(
        util.command(
            url="http://localhost:5000",
            json=True,
        ),
        stdout=subprocess.PIPE,
    )

    assert result.returncode == 0
    assert json.loads(result.stdout.decode()) == {
        "http://localhost:5000/bar": {
            "links": [
                {
                    "href": "/",
                    "url": "http://localhost:5000/",
                    "results": [
                        {
                            "type": "redirect",
                            "status_code": 302,
                            "value": "/foo",
                            "url": "http://localhost:5000/foo",
                        },
                        {
                            "type": "redirect",
                            "status_code": 302,
                            "value": "/bar",
                            "url": "http://localhost:5000/bar",
                        },
                        {
                            "type": "response",
                            "status_code": 200,
                        },
                    ],
                },
            ],
        },
    }


def test_json_not_found(http_server) -> None:
    http_server(blueprint=make_blueprint(last_redirect_ok=False), port=5000)

    result = subprocess.run(
        util.command(
            url="http://localhost:5000",
            json=True,
        ),
        stdout=subprocess.PIPE,
    )

    assert result.returncode == 1
    assert result.stdout.decode() == ""



from threading import Thread

import pytest
from flask import Blueprint
from werkzeug.serving import make_server

from . import server


@pytest.fixture
def http_server():
    created_servers = {}

    def _make_server(blueprint: Blueprint, port: int) -> None:
        app = server.create_app(blueprint=blueprint)
        srv = make_server("127.0.0.1", port, app)
        thread = Thread(target=srv.serve_forever)
        thread.daemon = True
        thread.start()
        created_servers[port] = (thread, srv)

    yield _make_server

    for thread, srv in created_servers.values():
        srv.shutdown()
        thread.join()



import subprocess

from . import util


def test_json() -> None:
    result = subprocess.run(
        util.command(url="http://localhost:1", json=True),
        stdout=subprocess.PIPE,
    )

    assert result.returncode == 1
    assert result.stdout.decode() == ""



from dataclasses import dataclass

from rich import print
from rich.markup import escape
from rich.tree import Tree

from . import analyzer, outcome


@dataclass(frozen=True)
class Converter(outcome.Converter[str]):
    def convert_page(self, page: outcome.Page) -> str:
        return f"[red]{escape(str(page.code))}[/red]"

    def convert_redirect(self, redirect: outcome.Redirect) -> str:
        return escape(str(redirect.code))

    def convert_request_error(self, error: outcome.RequestError) -> str:
        return f"[red]{escape(str(error.msg))}[/red]"

    def convert_excluded(self, excluded: outcome.Excluded) -> str:
        return "excluded"

    def convert_unknown(self, unknown: outcome.Unknown) -> str:
        raise AssertionError("`Unknown` result isn't supposed to be shown")


def print_results(analysis: analyzer.Analysis) -> None:
    failed_style = "bold dim" if analysis.ok() else "bold red"
    root_label = (
        f"📂 Results: [bold]{analysis.stats.total}[/bold] links"
        f" ([bold green]{analysis.stats.ok}[/bold green] ok,"
        f" [{failed_style}]{analysis.stats.failed}[/{failed_style}] failed)"
    )
    tree = Tree(root_label, guide_style="dim")

    for url, info in analysis.pages.items():
        bad_links = [link for link in info.links if not link.ok()]

        if not bad_links:
            continue

        branch = tree.add(f"📄 {escape(str(url))}", style="bold")

        for link in bad_links:
            items = link.results.convert_with(Converter())
            results = " → ".join(items)
            branch.add(f"🔗 [blue]{escape(link.href)}[/blue]: {results}")

    print(tree)



from typing import Sequence

import pytest

from discolinks.core import Url
from discolinks.html import get_hrefs, parse_href


@pytest.mark.parametrize(
    "body,expected",
    [
        ("<body></body>", []),
        ("""<a href="foo">""", ["foo"]),
        ("""<a href="foo#bar">""", ["foo#bar"]),
        ("""<a href="mailto:foo@example.net">""", ["mailto:foo@example.net"]),
        (
            """
            <?xml version="1.0" encoding="UTF-8" standalone="no"?>
            <svg>
                <a href="foo"></a>
            </svg>
            """,
            ["foo"],
        ),
    ],
)
def test_get_hrefs(body: str, expected: Sequence[str]):
    result = get_hrefs(body=body)

    assert list(result) == expected


@pytest.mark.parametrize(
    "href,base_url,expected",
    [
        (
            "",
            Url.from_str("http://example.net"),
            Url.from_str("http://example.net"),
        ),
        (
            "foo",
            Url.from_str("http://example.net"),
            Url.from_str("http://example.net/foo"),
        ),
        (
            "bar",
            Url.from_str("http://example.net/foo"),
            Url.from_str("http://example.net/bar"),
        ),
        (
            "bar",
            Url.from_str("http://example.net/foo/"),
            Url.from_str("http://example.net/foo/bar"),
        ),
        (
            "/bar",
            Url.from_str("http://example.net/foo"),
            Url.from_str("http://example.net/bar"),
        ),
        (
            "/bar",
            Url.from_str("http://example.net/foo/"),
            Url.from_str("http://example.net/bar"),
        ),
        (
            "http://example.org",
            Url.from_str("http://example.net"),
            Url.from_str("http://example.org"),
        ),
        (
            "//example.org",
            Url.from_str("http://example.net"),
            Url.from_str("http://example.org"),
        ),
        (
            "mailto:foo@example.net",
            Url.from_str("http://example.net"),
            None,
        ),
        (
            "tel:+0123",
            Url.from_str("http://example.net"),
            None,
        ),
    ],
)
def test_parse_href(href: str, base_url: Url, expected: Url):
    result = parse_href(href=href, base_url=base_url)

    assert result == expected



from dataclasses import dataclass
from typing import Mapping, Sequence

from . import outcome
from .core import Url
from .url_store import UrlInfo


@dataclass(frozen=True)
class LinkResult:
    href: str
    url: Url
    results: outcome.Results

    def ok(self) -> bool:
        return self.results.ok()


@dataclass(frozen=True)
class Page:
    links: Sequence[LinkResult]


def make_chain(
    url_infos: Mapping[Url, UrlInfo],
    start_url: Url,
) -> Sequence[outcome.Result]:
    """
    Build and return a redirect chain of URL results.

    Starting with a given URL, this follows redirects to determine the path leading to a
    web page, a connection error or a cancellation.
    """

    url = start_url
    chain: list[outcome.Result] = []

    while True:
        info = url_infos.get(url)

        if info is None:
            chain.append(outcome.Unknown())
            break

        result = info.result
        chain.append(result)
        redirect_url = result.redirect_url()

        if redirect_url is None:
            break
        else:
            url = redirect_url

    return chain


@dataclass
class Stats:
    ok: int = 0
    failed: int = 0

    @property
    def total(self) -> int:
        return self.ok + self.failed

    def add(self, ok: bool) -> None:
        if ok:
            self.ok += 1
        else:
            self.failed += 1


@dataclass(frozen=True)
class Analysis:
    stats: Stats
    pages: Mapping[Url, Page]

    def ok(self) -> bool:
        return self.stats.failed == 0


def analyze(url_infos: Mapping[Url, UrlInfo]) -> Analysis:
    """
    Analyze URL data obtained from web scraping.
    """

    pages = {}
    stats = Stats()

    for url, info in url_infos.items():
        if info.links is None:
            continue

        links = []

        for link in info.links:
            result = LinkResult(
                href=link.href,
                url=link.url,
                results=outcome.Results(
                    chain=make_chain(url_infos=url_infos, start_url=link.url)
                ),
            )
            stats.add(ok=result.ok())
            links.append(result)

        pages[url] = Page(links=links)

    return Analysis(pages=pages, stats=stats)



import asyncio
import functools
import logging
import signal
from typing import Optional
from urllib.parse import urldefrag, urlparse

import click
import rich.console
import rich.markup
from rich.logging import RichHandler

from . import analyzer, export, text
from .core import Url
from .excluder import Excluder, ExcluderRegexError
from .link_extractor import get_links
from .monitor import Monitor, new_monitor
from .requester import Requester
from .url_store import UrlInfo, UrlStore
from .worker import work

logger = logging.getLogger(__name__)


def parse_url_arg(url: str) -> Optional[Url]:
    """
    Parse a string representing a URL.

    This is intended for the URL provided as argument to the program. For example, if the
    scheme is missing, it assumes the user meant to use the `http` scheme, like many other
    programs using web URLs.
    """

    if not url.startswith("http://") and not url.startswith("https://"):
        url = f"http://{url}"

    parsed = urlparse(url)

    if not parsed.netloc:
        return None

    (url, _) = urldefrag(url)
    return Url.from_str(url)


async def find_links(
    max_parallel_requests: int,
    requester: Requester,
    url_store: UrlStore,
    monitor: Monitor,
    start_url: Url,
    first_urls: frozenset[Url],
) -> None:
    queue: asyncio.Queue[Url] = asyncio.Queue()

    for url in first_urls:
        queue.put_nowait(url)

    workers: list[asyncio.Task] = []

    for _ in range(max_parallel_requests):
        worker = asyncio.create_task(
            work(
                queue=queue,
                requester=requester,
                url_store=url_store,
                monitor=monitor,
                start_url=start_url,
            ),
        )
        workers.append(worker)

    # Wait for queue processing to finish, or for any worker to finish (which only happens
    # if that worker raised an exception).
    queue_task = asyncio.create_task(queue.join())
    await asyncio.wait([queue_task, *workers], return_when=asyncio.FIRST_COMPLETED)

    for worker in workers:
        worker.cancel()

    try:
        await asyncio.gather(*workers)  # ← worker exceptions raised here
    except asyncio.CancelledError:
        pass


async def main_async(
    max_parallel_requests: int,
    url_store: UrlStore,
    monitor: Monitor,
    excluder: Excluder,
    start_url: Url,
):
    requester = Requester(excluder=excluder)
    url = start_url

    while True:
        result = await requester.get(url)
        next_url = result.redirect_url()
        new_urls = url_store.add_page(
            url=url,
            info=UrlInfo(result=result, links=get_links(url=url, result=result)),
        )

        if next_url is None:
            break

        logger.info(f"Redirected to {next_url}")

        if next_url not in new_urls:
            logger.error("Detected circular redirects. Aborting.")
            exit(1)

        url = next_url

    error_msg = result.error_msg()
    if error_msg is not None:
        logger.error("%s", error_msg)
        exit(1)

    if not result.ok():
        logger.error("Bad response status code: %d", result.status_code())
        exit(1)

    await find_links(
        max_parallel_requests=max_parallel_requests,
        requester=requester,
        url_store=url_store,
        monitor=monitor,
        start_url=url,
        first_urls=new_urls,
    )


@click.command()
@click.option(
    "--verbose",
    is_flag=True,
    help="Increase verbosity to show debug messages.",
)
@click.option(
    "--max-parallel-requests",
    default=4,
    type=click.IntRange(min=1),
    help="Maximum of requests which can be in-flight at any given time.",
)
@click.option(
    "--json",
    "to_json",
    is_flag=True,
    help="Export results as JSON to the standard output.",
)
@click.option(
    "--exclude",
    default=[],
    type=str,
    multiple=True,
    help="""
        Regular expression matching URLs to exclude from the crawl.
        Can be supplied multiple times.
    """,
)
@click.option("--url", required=True, help="URL where crawling will start.")
@click.version_option(
    prog_name="discolinks",
    message="%(prog)s version %(version)s",
)
def main(
    verbose: bool,
    max_parallel_requests: int,
    to_json: bool,
    exclude: tuple[str, ...],
    url: str,
) -> None:
    console = rich.console.Console(stderr=True)
    main_logger = logging.getLogger("discolinks")
    level = logging.DEBUG if verbose else logging.INFO
    main_logger.setLevel(level)
    main_logger.addHandler(RichHandler(console=console, show_time=False))

    start_url = parse_url_arg(url)

    if start_url is None:
        logger.error("Invalid URL: %s", url)
        exit(1)

    url_store = UrlStore()

    try:
        excluder = Excluder.from_regexes(regexes=exclude)
    except ExcluderRegexError as error:
        logger.error(
            "Invalid value `[yellow]%s[/yellow]` for --exclude: %s",
            rich.markup.escape(error.regex),
            rich.markup.escape(error.msg),
            extra={"markup": True},
        )
        exit(1)

    try:
        with new_monitor(console=console) as monitor:
            # Set event loop
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            # Define main task (wrap in a future to make it cancellable).
            main_task = asyncio.ensure_future(
                main_async(
                    max_parallel_requests=max_parallel_requests,
                    url_store=url_store,
                    monitor=monitor,
                    excluder=excluder,
                    start_url=start_url,
                )
            )

            # Cancel main task when interrupted.
            loop.add_signal_handler(
                signal.SIGINT,
                functools.partial(main_task.cancel, msg="SIGINT"),
            )
            loop.add_signal_handler(
                signal.SIGTERM,
                functools.partial(main_task.cancel, msg="SIGTERM"),
            )

            # Run main task.
            loop.run_until_complete(main_task)
    except asyncio.CancelledError as error:
        logger.warning("Interrupted (%s)", error)
        interrupted = True
    except Exception as exc:
        logger.exception(exc)
        interrupted = True
    else:
        interrupted = False

    url_infos = url_store.get_url_infos()
    analysis = analyzer.analyze(url_infos)
    ok = not interrupted and analysis.ok()

    if to_json:
        print(export.dump_json(analysis=analysis))
    else:
        text.print_results(analysis=analysis)

    exit(0 if ok else 1)



import warnings
from typing import Optional, Sequence
from urllib.parse import urldefrag, urljoin, urlparse

import bs4

from .core import Link, Url


def get_hrefs(body: str) -> Sequence[str]:
    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore", category=bs4.XMLParsedAsHTMLWarning)
        soup = bs4.BeautifulSoup(body, features="html.parser")

    return [url for a in soup.find_all("a") if (url := a.attrs.get("href")) is not None]


def parse_href(href: str, base_url: Url) -> Optional[Url]:
    """
    Parse the value of an `href` HTML attribute into a URL.

    If the link is relative, we need the base URL to infer the absolute URL.
    """

    (href, _) = urldefrag(href)
    parsed = urlparse(href)

    if parsed.scheme not in ("", "http", "https"):
        return None

    if parsed.netloc:
        if not parsed.scheme and href.startswith("//"):
            url = f"http:{href}"
        else:
            url = href
    else:
        url = urljoin(base_url.full, href)

    return Url.from_str(url)


def get_links(body: str, url: Url) -> Sequence[Link]:
    return [
        Link(href=href, url=link_url)
        for href in get_hrefs(body)
        if (link_url := parse_href(href, base_url=url)) is not None
    ]
